@@ -188,6 +188,7 @@ gboolean	camel_imapx_server_refresh_info_sync
 gboolean	camel_imapx_server_sync_changes_sync
 						(CamelIMAPXServer *is,
 						 CamelIMAPXMailbox *mailbox,
+						 gboolean can_influence_flags,
 						 GCancellable *cancellable,
 						 GError **error);
 gboolean	camel_imapx_server_expunge_sync	(CamelIMAPXServer *is,
@@ -272,6 +273,11 @@ GPtrArray *	camel_imapx_server_uid_search_sync
 						 const gchar * const *words,
 						 GCancellable *cancellable,
 						 GError **error);
+gboolean	camel_imapx_server_can_use_idle	(CamelIMAPXServer *is);
+gboolean	camel_imapx_server_is_in_idle	(CamelIMAPXServer *is);
+CamelIMAPXMailbox *
+		camel_imapx_server_ref_idle_mailbox
+						(CamelIMAPXServer *is);
 gboolean	camel_imapx_server_schedule_idle_sync
 						(CamelIMAPXServer *is,
 						 CamelIMAPXMailbox *mailbox,
